/************************************************************************
 *
 * All dialogs opened are created and used modal.
 *
 ************************************************************************
 * (C) Craig Drummond, 2006
 ************************************************************************
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 ************************************************************************/

#define _GNU_SOURCE
#include <dlfcn.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/wait.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <pwd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <qstring.h>
#include <qstringlist.h>
#include <qwidget.h>
#include <qapplication.h>
#include <qfiledialog.h>
#include <qthread.h>
#include <qnamespace.h>
#include <qeventloop.h>
#include "connect.h"

static bool useKde=false;

static bool writeString(int fd, const QString &str)
{
    int size=str.length()+1;

    return writeBlock(fd, (char *)&size, 4) && writeBlock(fd, str.local8Bit().data(), size);
}

static bool writeBool(int fd, bool b)
{
    char bv=b ? 1 : 0;

    return writeBlock(fd, (char *)&bv, 1);
}

class KQtThread : public QThread
{
    public:

    KQtThread(QStringList &l, QString &s, int f) : kdialogdError(false), res(l), selFilter(s), fd(f) { }

    bool readData(QCString &buffer, int size)
    {
        buffer.resize(size);
        return ::readBlock(fd, buffer.data(), size);
    }

    bool readString(QCString &str)
    {
        int size;
    
        if(!::readBlock(fd, (char *)&size, 4))
            return false;

        str.resize(size);
        return readBlock(fd, str.data(), size);
    }

    void run()
    {
        QCString buffer;
        int      num=0;

        if(readBlock(fd, (char *)&num, 4))
        {
            int n;

            for(n=0; n<num && !kdialogdError; ++n)
            {
                int size=0;

                if(readBlock(fd, (char *)&size, 4))
                {
                    if(size>0)
                    {
                        if(readData(buffer, size))
                        {
                            buffer[size-1]='\0';
                            if('/'==buffer[0])
                                res.append(buffer);
                            else
                                selFilter=buffer;
                        }
                        else
                            kdialogdError=true;
                    }
                }
                else
                    kdialogdError=true;
            }
        }
        else
            kdialogdError=true;
    }

    bool        kdialogdError;
    QStringList &res;
    QString     &selFilter;
    int         fd;
};

static bool sendMessage(QWidget *parent, Operation op, QStringList &res, QString &selFilter, const QString &title,
                        const char *p1, const char *p2, bool ow)
{
    if(connectToKDialogD())
    {
        char o=(char)op;
        int  xid=parent ? parent->topLevelWidget()->winId() : qApp->activeWindow()->winId();

        if(writeBlock(kdialogdSocket, &o, 1) &&
           writeBlock(kdialogdSocket, (char *)&xid, 4) &&
           writeString(kdialogdSocket, title) &&
           (p1 ? writeString(kdialogdSocket, p1) : true) &&
           (p2 ? writeString(kdialogdSocket, p2) : true) &&
           (OP_FILE_SAVE==op ? writeBool(kdialogdSocket, ow) : true))
        {
            KQtThread thread(res, selFilter, kdialogdSocket);

            thread.start();
            while(thread.running())
                QApplication::eventLoop()->processEvents(QEventLoop::ExcludeUserInput, 10);

            if(thread.kdialogdError)
            {
                closeConnection();
                return false;
            }
            return true;
        }
    }

    return false;
}

static QString getTitle(const QString &title, Operation op)
{
    if(!title.isEmpty())
        return title;

    switch(op)
    {
        case OP_FILE_OPEN:
            return "Open File";
        case OP_FILE_OPEN_MULTIPLE:
            return "Open File(s)";
        case OP_FILE_SAVE:
            return "Save File";
        case OP_FOLDER:
            return "Choose Folder";
        default:
            return NULL;
    }
}

static bool openKdeDialog(QWidget *widget, const QString &title, const char *p1, const char *p2, Operation op,
                          QStringList &res, QString *selFilter, bool ow=false)
{
    QString filter;
    bool    rv=sendMessage(widget, op, res, filter, getTitle(title, op), p1, p2, ow);

    /* If we failed to talk to, or start kdialogd, then dont keep trying - just fall back to Gtk */
    if(!rv)
        useKde=false;
    else if(selFilter)
        *selFilter=filter;

    return rv;
}

static void kgtkExit()
{
    if(useKde)
        closeConnection();
}

static bool kgtkInit()
{
    static bool initialised=false;

    if(!initialised)
    {
        initialised=true;
        useKde=NULL!=getenv("KDE_FULL_SESSION") && connectToKDialogD();
        if(useKde)
            atexit(&kgtkExit);
    }

    return useKde;
}

static QString fixFilter(const QString &f)
{
    QString               filter;
    QTextOStream          str(&filter);
    QStringList           list(QStringList::split(";;", f));
    QStringList::Iterator it(list.begin()),
                          end(list.end());
    bool                  first=true;

    for(; it!=end; ++it)
    {
        int ob=(*it).findRev('('),
            cb=(*it).findRev(')');

        if(-1!=cb && ob<cb)
        {
            if(first)
                first=false;
            else
                str << '\n';
            str << (*it).mid(ob+1, (cb-ob)-1) << '|' << (*it).mid(0, ob);
        }
    }

    return filter;
}

QString QFileDialog::getOpenFileName(const QString &initially, const QString &filter,  QWidget *parent,
                                     const char *name, const QString &caption, QString *selectedFilter,
                                     bool resolveSymlinks)
{
    QStringList res;
    QString     f(fixFilter(filter));
    kgtkInit();

    return openKdeDialog(parent, caption, initially.isEmpty() ? "" : initially.local8Bit(),
                         f.isEmpty() ? "" : f.local8Bit(), OP_FILE_OPEN, res, selectedFilter)
            ? res.first()
            : QString::null;
}

QString QFileDialog::getSaveFileName(const QString &initially, const QString &filter, QWidget *parent,
                                     const char *name, const QString &caption, QString *selectedFilter,
                                     bool resolveSymlinks)
{
    QStringList res;
    QString     f(fixFilter(filter));
    kgtkInit();

    return openKdeDialog(parent, caption, initially.isEmpty() ? "" : initially.local8Bit(),
                         f.isEmpty() ? "" : f.local8Bit(), OP_FILE_SAVE, res, selectedFilter)
            ? res.first()
            : QString::null;
}

QString QFileDialog::getExistingDirectory(const QString &dir, QWidget *parent, const char *name,
                                          const QString &caption, bool dirOnly, bool resolveSymlinks)
{
    QStringList res;
    kgtkInit();

    return openKdeDialog(parent, caption, dir.isEmpty() ? "" : dir.local8Bit(), NULL, OP_FOLDER, res, NULL)
            ? res.first()
            : QString::null;
}

QStringList QFileDialog::getOpenFileNames(const QString &filter, const QString &dir, QWidget *parent,
                                          const char *name, const QString &caption, QString *selectedFilter,
                                          bool resolveSymlinks)
{
    QStringList res;
    QString     f(fixFilter(filter));
    kgtkInit();

    openKdeDialog(parent, caption, dir.isEmpty() ? "" : dir.local8Bit(), f.isEmpty() ? "" : f.local8Bit(),
                  OP_FILE_OPEN_MULTIPLE, res, selectedFilter);
    return res;
}
